/* stmt.c
	P. Chongstitvatana				25 Aug 1997

	som-v16, i-array				6 Nov 2005
	som-v17 newgvar					24 Dec 2005
	som-v23 static array with str	28 Jan 2006 (Chinese New Year)
*/

#include "compile.h"

extern int lv;		  	// defined in symtab.c

int GV = 65;			// gvar 64..255, 64 is retval
int nenum = 0;		 	// enum number

/* ========= push down stack ======== */

PRIVATE int ystack[MAXYSTK];
PRIVATE int ysp = 0;

void ypush(int item){
	ysp++;
	if(ysp >= MAXYSTK)
		seterror("parser stack overflow");
	ystack[ysp] = item;
}

int ypop(void){
	int x;
	if(ysp <= 0)
		seterror("parser stack underflow");
	x = ystack[ysp];
	ysp--;
	return x;
}

int ytos(void){	return ystack[ysp];}
//int getysp(void){ return ysp; }

/* ============================= */

void expect( int tktype, char *mess ){
  if( tok != tktype ) seterror(mess);
}

void commit( int status )  {
  if ( status == 0 ) seterror("syntax");
}

// enter function/macro name into symbol table
// [ -- name]
void setfname(int ty){
	int idx;
	char mess[80];
	idx = search(tokstring);
	if(getType(idx) == tyNEW)
		setType(idx,ty);
	else
		warning("redefine",tokstring);
	ypush(idx);
}

// [name arity -- name]
void setarity(void){
	int arity, idx;
	arity = ypop();
	idx = ytos();
	setArg(idx,arity);
	setLv(idx,lv);
}

// [name %ex -- %to]
void dofun(int op){
	int idx, e, a;
	e = ypop();
	idx = ypop();
	#ifdef VERBOSE
		printf("%s\n",getName(idx));
	#endif
	a = cons(newatom(GNAME,idx),list(e));
	ypush(cons(newatom(OPER,op),a));
	lv = 0;				// clear local symbol
}

// [%lv %ex0 %ex0 %ex -- %for]  (for lv ex0 ex0 ex)
void dofor(void){
	int e;
	e = ypop();
	e = cons(ypop(),list(e));
	e = cons(ypop(),e);
	e = cons(ypop(),e);
	ypush(cons(newatom(OPER,tkFOR),e));
}

// [%ex0 %ex -- %while] (while ex0 ex)
void dowhile(void){
	int e;
	e = ypop();
	e = cons(ypop(),list(e));
	ypush(cons(newatom(OPER,tkWHILE),e));
}

// [%ex0 %ex [%ex MARK] -- %if/ifelse ]
void doif(void){
	int e;
	if( ytos() == MARK ){
		e = ypop();  		// throw away MARK
		e = ypop();
		e = cons(ypop(),list(e));
		e = cons(ypop(),e);	// (ex0 ex ex)
		ypush(cons(newatom(OPER,tkELSE),e));
	}else{
		e = ypop();
		e = cons(ypop(),list(e));
		ypush(cons(newatom(OPER,tkIF),e));
	}
}

// globals are stored in M[]
//	global grows to end
int newdata(void){
	int a;
	a = DS;
	DS++;
	if( DS >= MAXDS )
		seterror("out of memory");
	return a;
}

PRIVATE int newgvar(void){
	int a;
	a = GV;
	GV++;
	if(GV > MAXGVAR)
		seterror("too many gvar");
	return a;
}

// [ -- %var]
PRIVATE void dovar(int idx){
	int type;
	type = getType(idx);
	switch(type){
	case tyGVAR:
		ypush(newatom(GNAME,idx));
		break;
	case tyLOCAL:
		ypush(newatom(LNAME,getRef(idx)));
		break;
	default:
		seterror("expect variable");
	}
}

// parse index var (local) of for loop
void dolocal(void){
	int idx, type, var;
	idx = search(tokstring);
	if( getType(idx) == tyLOCAL)
		dovar(idx);
	else
		seterror("for: expect local var");
}

// [ %var %ex -- %vec]  (vec var ex)
void dovec(void){
	int e, v;
	e = ypop();
	v = ypop();
	if( isatom(v) )
		ypush(cons(newatom(OPER,tkLBRACKET),cons(v,list(e))));
	else
		seterror("expect variable");
}

// [ %var %ex -- %assign] (= var ex)
PRIVATE void doset(void){
	int e;
	e = ypop();
	ypush(cons(newatom(OPER,tkEQ),cons(ypop(),list(e))));
}

// left variable,
//   if IDEN: NEW, GVAR, LOCAL then dovar(), mod()
//   else return 0
PRIVATE int lval(void){
	int idx, type, ref;
	if( tok != tkIDEN ) return 0;
	idx = search(tokstring);
	type = getType(idx);
	switch(type){
	case tyNEW:
		ref = newgvar();	// new gvar, alloc DS
		setType(idx,tyGVAR);
		setRef(idx,ref);
		break;
	case tyGVAR:
	case tyLOCAL: break;
	default: return 0;
	}
	dovar(idx);
	lex();
	commit(mod());
	return 1;
}
// combine exas with ex0
// if assignment, [-- %var/%vec]
int exas(void){
	if( lval() ){
		if( tok == tkEQ ){
			lex();    // skip =
			commit(ex0());
			doset();
		}else {
			// lval() has already done
			// half of the job, term()
			commit(terms());
		}
		return 1;
	}
	return ex0();
}

void doenum(void){
	int idx;
	idx = search(tokstring);
	if(getType(idx) == tyNEW){
		setType(idx,tyENUM);
		setRef(idx,nenum);
  		nenum++;
	}else
		seterror("enum: expect unique label");
}

// function/macro call, parse ex0^arity
// [-- %call] (fun name ex... ex)
PRIVATE void docall(int idx, int op){
	int i, arg, e;
	arg = getArg(idx);
	for(i=0; i<arg; i++)
		commit( ex0() ); 			   	// parse ex^arity
	e = NIL;
	for(i=0; i<arg; i++)
		e = cons(ypop(),e);				// (ex ... ex)
	e = cons(newatom(GNAME,idx),e);		// (name ex .. ex)
	ypush(cons(newatom(OPER,op),e));	// (fun/mac name ex .. ex)
}

// [ -- %num]
void donum(int value){
	ypush(newatom(NUM,value));
}

// already lex() the next token
// ID is : var, fun-call, enum
void doiden(int idx){
	switch(getType(idx)){
	case tyNEW:
//		seterror("undefined variable");
		warning(getName(idx),"new global");
		setType(idx,tyGVAR);
		setRef(idx,newgvar());
		dovar(idx);
		commit(mod());
		break;
	case tyGVAR:
	case tyLOCAL:
		dovar(idx);
		commit(mod());
		break;
	case tyFUNC:
		docall(idx,tkBE);
		break;
	case tyMAC:
	case tyFULL:
		docall(idx,tkRBRACKET);
		break;
	case tyENUM:
		donum(getRef(idx));
		break;
	}
}

// [%ex1 %ex2 -- %bop]  (bop ex1 ex2)
void dobop(int bop){
	int ex1, ex2, a;
	ex2 = ypop();
	ex1 = ypop();
	a = cons(ex1, list(ex2));
	ypush(cons(newatom(OPER,bop),a));
}

// block = tkBB, simplify block size 0 and 1
PRIVATE int makeblock(int a){
	if( a == NIL )
		return NIL;				// block size 0 {} => NIL
	else if( cdr(a) == NIL )
	    return car(a);			// block size 1 {a} => a
	else
		return cons(newatom(OPER,tkBB),a);
}

// [MARK %ex1...%exn -- %block] (block ex1.. exn)
void doblock(void){
	int e, a;
	a = NIL;
	while( (e = ypop()) != MARK )
		a = cons(e,a);
	ypush(makeblock(a));
}

// [%ex -- %uop] (uop ex)
void douop(int uop){
	int e;
	e = ypop();
	ypush(cons(newatom(OPER,uop),list(e)));
}

// [ MARK %num %ex0 ... %ex0  -- %sys]
// (sys num ex0 ... ex0)
void dosys(void){
	int e, a;
	e = NIL;
	while( (a = ypop()) != MARK )
		e = cons(a,e);
	ypush(cons(newatom(OPER,tkSYSCALL),e));
}

// convert label to number [ -- %num ]
void dolabel(void){
	int idx, type;
	idx = search(tokstring);
	if(getType(idx) != tyENUM)
		seterror("expect label");
	donum(getRef(idx));
}

// [%ex0 MARK %num %ex ... %-1 %ex -- %case]
void docase(void){
	int e, a;
	a = NIL;
	while( (e = ypop()) != MARK ){
		e = cons(ypop(),list(e));	// e = (num ex)
		a = cons(e,a);
	}
	a = makeblock(a);
	e = cons(ypop(),list(a));		// (ex0 block)
	ypush(cons(newatom(OPER,tkCASE),e));
}
// unwind ystack, store values in M[.], a at the front
// values started at ystack[a+1]..ystack[ysp]
/*
void dostatic(int a){
	int i, p, y;
	y = getysp();
	p = xalloc(y - a);
	for(i=p+y-a-1; i>=p; i--){	// mov back to front
		M[i] = ypop();
		printf("%d:%d ",i,M[i]);
	}
	ypush(newatom(ADS,p));
}
*/
// convert tokstring to som-string and
// store it to memory
void akeepStr(char *s){
	int i;
	strpack(SSTRING,s);
	i = SSTRING;
	while( M[i] != 0 ){
		M[newdata()] = M[i];
		i++;
	}
	M[newdata()] = 0;
}
// identifier in static array
void akeepIden(char *s){
	int idx;
	idx = search(s);
	if(getType(idx) == tyNEW)
		seterror("unknown identifier");
	M[newdata()] = getRef(idx);
}

// ---------- string functions ----------

// pack C-string (char *) to S-string in M[] indexed by s1
void strpack(int s1, char *s2){
	uchar c1, c2, c3, c4;
	while( *s2 != 0 ){
		c1 = (*s2 != 0) ? *s2++ : 0;
		c2 = (*s2 != 0) ? *s2++ : 0;
		c3 = (*s2 != 0) ? *s2++ : 0;
		c4 = (*s2 != 0) ? *s2++ : 0;
		M[s1] = (((((((int)c1)<<8)|(int)c2)<<8)|
			(int)c3)<<8)|(int)c4;
		s1++;
	}
	M[s1] = 0;
}

// convert som-string to C-string
void strunpack(char *s1, int s2){
	int s;
	char c1, c2, c3, c4;
	while( M[s2] != 0 ){
		s = M[s2];
		c1 = (char)((s >> 24) & 255);
		c2 = (char)((s >> 16) & 255);
  		c3 = (char)((s >> 8) & 255);
   		c4 = (char)(s & 255);
   		if( c1 != 0 ) *s1++ = c1;
		if( c2 != 0 ) *s1++ = c2;
		if( c3 != 0 ) *s1++ = c3;
		if( c4 != 0 ) *s1++ = c4;
		s2++;
	}
	*s1 = 0;
}

// [ -- %string ]   string is an int pointed to heap
void dostring(char *s){
	int len, k, s2;
	len = strlen(s);
	k = len / 4;
	if( (len % 4) != 0 ) k++;
	s2 = xalloc(k+1);
	strpack(s2,s);
	ypush(newatom(STRING,s2));
}

void prstr(int i){
	int c1,c2,c3,c4;

	while( M[i] != 0 ){
		c1 = (M[i] >> 24) & 255;
		c2 = (M[i] >> 16) & 255;
		c3 = (M[i] >> 8) & 255;
		c4 = M[i] & 255;
		if( c1 != 0 ) printf("%c",c1);
		if( c2 != 0 ) printf("%c",c2);
		if( c3 != 0 ) printf("%c",c3);
		if( c4 != 0 ) printf("%c",c4);
		i++;
	}
}

/* -------------- DEBUG ---------------- */
/*
void teststr(){
	char s[] = "abcdef";
	int s2;
	printf("/%s/\n",s);
	s2 = 10;
	strpack(s2,s);
	prstr(s2);

	strpack(1,"a cow jumps over the moon");
	strunpack(name,1);
	printf("%s\n",name);
	exit(0);
}
*/
